
from cStringIO import StringIO
from repozitory.interfaces import IArchive
from repozitory.interfaces import IContainerRecord
from repozitory.interfaces import IDeletedItem
from repozitory.interfaces import IObjectHistoryRecord
from repozitory.schema import ArchivedBlobInfo
from repozitory.schema import ArchivedBlobLink
from repozitory.schema import ArchivedChunk
from repozitory.schema import ArchivedClass
from repozitory.schema import ArchivedContainer
from repozitory.schema import ArchivedCurrent
from repozitory.schema import ArchivedItem
from repozitory.schema import ArchivedItemDeleted
from repozitory.schema import ArchivedObject
from repozitory.schema import ArchivedState
from repozitory.schema import Base
from sqlalchemy import func
from sqlalchemy.engine import create_engine
from sqlalchemy.orm import scoped_session
from sqlalchemy.orm.session import sessionmaker
from zope.interface import implements
from zope.sqlalchemy import ZopeTransactionExtension
import datetime
import hashlib
import tempfile

_global_sessions = {}  # {db_string: SQLAlchemy session}


def forget_sessions():
    _global_sessions.clear()


class EngineParams(object):
    """Parameters to pass to SQLAlchemy's create_engine() call.

    db_string is an URL such as postgresql://localhost:5432/ .
    The keyword parameters are documented here:

    http://www.sqlalchemy.org/docs/core/engines.html#sqlalchemy.create_engine
    """
    def __init__(self, db_string, **kwargs):
        self.db_string = db_string
        self.kwargs = kwargs


def unicode_or_none(s):
    return unicode(s) if s is not None else None


def find_class(module, name):
    m = __import__(module, None, None, ('__doc__',))
    return getattr(m, name, None)


class Archive(object):
    """An object archive that uses SQLAlchemy.

    Note: instances of this class may be stored in ZODB, so instances
    must only hold things that can be pickled.
    """
    implements(IArchive)

    chunk_size = 1048576    # Store blobs in chunks of this size

    def __init__(self, engine_params):
        self.engine_params = engine_params

    @property
    def session(self):
        """Get the SQLAlchemy session.  Uses a global session pool."""
        params = self.engine_params
        db_string = params.db_string
        session = _global_sessions.get(db_string)
        if session is None:
            engine = create_engine(db_string, **params.kwargs)
            session = self._create_session(engine)
            _global_sessions[db_string] = session
        return session

    def _create_session(self, engine):
        Base.metadata.create_all(engine)
        # Distinguish sessions by thread.
        session = scoped_session(sessionmaker(
            extension=ZopeTransactionExtension()))
        session.configure(bind=engine)
        return session

    def archive(self, obj):
        """Add a version to the archive of an object.

        The object does not need to have been in the archive
        previously.  The object must provide the IObjectVersion interface.

        Returns the new version number.
        """
        docid = obj.docid
        session = self.session
        max_version = None
        arc_obj = (session.query(ArchivedObject)
            .filter_by(docid=docid)
            .first())
        if arc_obj is None:
            arc_obj = ArchivedObject(
                docid=docid,
                created=obj.created,
            )
            session.add(arc_obj)
        else:
            max_version = (
                session.query(func.max(ArchivedState.version_num))
                .filter_by(docid=docid)
                .scalar())

        arc_current = (
            session.query(ArchivedCurrent)
            .filter_by(docid=docid)
            .first())
        if arc_current is None:
            derived_from_version = None
        else:
            derived_from_version = arc_current.version_num

        klass = getattr(obj, 'klass', None)
        if klass is None:
            klass = obj.__class__
        class_id = self._prepare_class_id(klass)

        arc_state = ArchivedState(
            docid=docid,
            version_num=(max_version or 0) + 1,
            derived_from_version=derived_from_version,
            archive_time=datetime.datetime.utcnow(),
            class_id=class_id,
            path=unicode(obj.path),
            modified=obj.modified,
            user=unicode(obj.user),
            title=unicode_or_none(obj.title),
            description=unicode_or_none(obj.description),
            attrs=obj.attrs,
            comment=unicode_or_none(obj.comment),
        )
        session.add(arc_state)

        blobs = getattr(obj, 'blobs', None)
        if blobs:
            for name, value in blobs.items():
                self._link_blob(arc_state, name, value)

        arc_current = (
            session.query(ArchivedCurrent)
            .filter_by(docid=docid)
            .first())
        if arc_current is None:
            arc_current = ArchivedCurrent(
                docid=docid,
                version_num=arc_state.version_num,
            )
            session.add(arc_current)
        else:
            arc_current.version_num = arc_state.version_num
        session.flush()
        return arc_state.version_num

    def _prepare_class_id(self, klass):
        """Add a class or reuse an existing class ID."""
        session = self.session
        module = unicode(klass.__module__)
        name = unicode(klass.__name__)
        actual = find_class(module, name)
        if actual != klass:
            raise TypeError("Broken class reference: %s != %s" % (
                actual, klass))
        cls = (session.query(ArchivedClass)
            .filter_by(module=module, name=name)
            .first())
        if cls is None:
            cls = ArchivedClass(module=module, name=name)
            session.add(cls)
            session.flush()
        return cls.class_id

    def _link_blob(self, arc_state, name, value):
        """Link a named blob to an object state."""
        if isinstance(value, basestring):
            fn = value
            f = open(fn, 'rb')
            try:
                blob_id = self._prepare_blob_id(f)
            finally:
                f.close()
        else:
            f = value
            f.seek(0)
            blob_id = self._prepare_blob_id(f)

        a = ArchivedBlobLink(
            docid=arc_state.docid,
            version_num=arc_state.version_num,
            name=unicode(name),
            blob_id=blob_id,
        )
        arc_state.blob_links.append(a)

    def _prepare_blob_id(self, f):
        """Upload a blob or reuse an existing blob containing the same data."""
        # Compute the length and hashes of the blob data.
        length = 0
        md5_calc = hashlib.md5()
        sha256_calc = hashlib.sha256()
        while True:
            data = f.read(self.chunk_size)
            if not data:
                break
            length += len(data)
            md5_calc.update(data)
            sha256_calc.update(data)
        md5 = md5_calc.hexdigest()
        sha256 = sha256_calc.hexdigest()
        f.seek(0)

        session = self.session
        arc_blob = (
            session.query(ArchivedBlobInfo)
            .filter_by(length=length, md5=md5, sha256=sha256)
            .first())
        if arc_blob is not None:
            return arc_blob.blob_id

        arc_blob = ArchivedBlobInfo(
            chunk_count=0,
            length=length,
            md5=md5,
            sha256=sha256,
        )
        session.add(arc_blob)
        session.flush()  # Assign arc_blob.blob_id
        blob_id = arc_blob.blob_id

        # Upload the data.
        chunk_index = 0
        while True:
            data = f.read(self.chunk_size)
            if not data:
                break
            arc_chunk = ArchivedChunk(
                blob_id=blob_id,
                chunk_index=chunk_index,
                chunk_length=len(data),
                data=data,
            )
            arc_blob.chunks.append(arc_chunk)
            session.flush()
            chunk_index += 1

        arc_blob.chunk_count = chunk_index
        session.flush()
        return arc_blob.blob_id

    def history(self, docid, only_current=False):
        """Get the history of an object.

        Returns a list of IObjectHistoryRecord.
        The most recent version is listed first.
        """
        created = (self.session.query(ArchivedObject.created)
            .filter_by(docid=docid)
            .scalar())
        current_version = (self.session.query(ArchivedCurrent.version_num)
            .filter_by(docid=docid)
            .scalar())
        q = self.session.query(ArchivedState).filter_by(docid=docid)
        if only_current:
            q = q.filter_by(version_num=current_version)
        rows = q.order_by(ArchivedState.version_num.desc()).all()
        return [ObjectHistoryRecord(row, created, current_version)
            for row in rows]

    def get_version(self, docid, version_num):
        """Return a specific IObjectHistoryRecord for a document.
        """
        created = (self.session.query(ArchivedObject.created)
            .filter_by(docid=docid)
            .scalar())
        current_version = (self.session.query(ArchivedCurrent.version_num)
            .filter_by(docid=docid)
            .scalar())
        row = (self.session.query(ArchivedState)
            .filter_by(docid=docid, version_num=version_num)
            .one())
        return ObjectHistoryRecord(row, created, current_version)

    def reverted(self, docid, version_num):
        """Tell the database that an object has been reverted."""
        session = self.session
        row = session.query(ArchivedCurrent).filter_by(docid=docid).one()
        row.version_num = version_num
        session.flush()

    def archive_container(self, container, user):
        """Update the archive of a container.

        The container parameter must provide the IContainerVersion interface.
        The container does not need to have been in the archive previously.

        Note that this method does not keep a record of container versions.
        It only records the contents of the container and tracks deletions,
        to allow for undeletion of contained objects.

        Returns None.
        """
        container_id = container.container_id
        session = self.session
        now = datetime.datetime.utcnow()
        arc_container = (session.query(ArchivedContainer)
            .filter_by(container_id=container_id)
            .first())
        path = unicode(container.path)
        if arc_container is None:
            arc_container = ArchivedContainer(
                container_id=container_id,
                path=path,
            )
            session.add(arc_container)
            item_list = []
            deleted_list = []
        else:
            if arc_container.path != path:
                arc_container.path = path
            item_list = (session.query(ArchivedItem)
                .filter_by(container_id=container_id)
                .all())
            deleted_list = (session.query(ArchivedItemDeleted)
                .filter_by(container_id=container_id)
                .all())

        old_items = {}  # {(ns, name): ArchivedItem}
        old_docid_names = {}  # {docid: (ns, name)}
        for item in item_list:
            k = (item.namespace, item.name)
            old_items[k] = item
            old_docid_names[item.docid] = k

        new_items = {}  # {(ns, name): docid}
        current_docids = set()
        if container.map:
            for name, docid in container.map.items():
                new_items[(u'', unicode(name))] = docid
                current_docids.add(docid)
        if container.ns_map:
            for ns, m in container.ns_map.items():
                ns = unicode(ns)
                for name, docid in m.items():
                    new_items[(ns, unicode(name))] = docid
                    current_docids.add(docid)

        for k in set(new_items).difference(old_items):
            # Add an item to the container.
            ns, name = k
            docid = new_items[k]
            item = ArchivedItem(
                container_id=container_id,
                namespace=ns,
                name=name,
                docid=docid,
            )
            session.add(item)

        for k in set(old_items).difference(new_items):
            # Remove an item from the container.
            item = old_items[k]
            session.delete(item)

        for k in set(old_items).intersection(new_items):
            item = old_items[k]
            docid = new_items[k]
            if item.docid != docid:
                # An item changed its docid.
                item.docid = docid

        for row in deleted_list:
            if row.docid in current_docids:
                # This item exists, so remove the deletion record.
                session.delete(row)

        for docid in set(old_docid_names).difference(current_docids):
            # At least one item has just been deleted.
            ns, name = old_docid_names[docid]
            row = ArchivedItemDeleted(
                container_id=container_id,
                docid=docid,
                namespace=ns,
                name=name,
                deleted_time=now,
                deleted_by=unicode(user),
            )
            session.add(row)

    def container_contents(self, container_id):
        """Return the contents of a container as IContainerRecord.
        """
        session = self.session
        row = (session.query(ArchivedContainer)
            .filter_by(container_id=container_id)
            .one())
        return ContainerRecord(self.session, row)


class ObjectHistoryRecord(object):
    implements(IObjectHistoryRecord)

    _blobs = None
    _klass = None

    def __init__(self, state, created, current_version):
        self._state = state
        self.current_version = current_version
        self.derived_from_version = state.derived_from_version
        self.created = created
        self.modified = state.modified
        self.title = state.title
        self.description = state.description
        self.docid = state.docid
        self.path = state.path
        self.attrs = state.attrs or {}
        self.version_num = state.version_num
        self.archive_time = state.archive_time
        self.user = state.user
        self.comment = state.comment

    @property
    def blobs(self):
        blobs = self._blobs
        if blobs is None:
            blobs = {}
            for link in self._state.blob_links:
                blobs[link.name] = BlobReader(link.blob)
            self._blobs = blobs
        return blobs

    @property
    def klass(self):
        res = self._klass
        if res is None:
            cls = self._state.class_
            self._klass = res = find_class(cls.module, cls.name)
        return res


class BlobReader(object):
    """Reads a blob file on demand and delegates to the open file."""

    _file = None
    _max_stringio = 1048576  # If blobs are larger than this, use a temp file.

    def __init__(self, blob):
        self._blob = blob

    def _get_file(self):
        f = self._file
        if f is None:
            length = self._blob.length
            if length <= self._max_stringio:
                # The blob fits in memory.
                f = StringIO()
            else:
                # Write the blob to a temporary file.
                f = tempfile.TemporaryFile()
            for chunk in self._blob.chunks:
                f.write(chunk.data)
            f.seek(0)
            self._file = f
        return f

    def __getattr__(self, name):
        return getattr(self._get_file(), name)

    def write(self, data):
        raise IOError("BlobReader is not writable")

    def writelines(self, data):
        raise IOError("BlobReader is not writable")


class ContainerRecord(object):
    implements(IContainerRecord)

    def __init__(self, session, row):
        self.container_id = row.container_id
        self.path = row.path

        self.map = {}
        self.ns_map = {}
        item_list = (session.query(ArchivedItem)
            .filter_by(container_id=self.container_id)
            .all())
        for item in item_list:
            ns = item.namespace
            name = item.name
            if ns:
                m = self.ns_map.get(ns)
                if m is None:
                    self.ns_map[ns] = m = {}
                m[name] = item.docid
            else:
                self.map[name] = item.docid

        deleted_rows = (session.query(ArchivedItemDeleted)
            .filter_by(container_id=self.container_id)
            .order_by(ArchivedItemDeleted.deleted_time.desc(),
                ArchivedItemDeleted.namespace, ArchivedItemDeleted.name)
            .all())
        new_container_map = {}  # {docid: [new_container_id]}
        if deleted_rows:
            # Get the list of new container_ids for all objects
            # deleted from this container.
            docids = [row.docid for row in deleted_rows]
            new_container_rows = (
                session.query(ArchivedItem.docid, ArchivedItem.container_id)
                .filter(ArchivedItem.docid.in_(docids))
                .all()
            )
            for docid, container_id in new_container_rows:
                new_container_map.setdefault(docid, []).append(container_id)
        self.deleted = [DeletedItem(row, new_container_map.get(row.docid))
            for row in deleted_rows]


class DeletedItem(object):
    implements(IDeletedItem)

    def __init__(self, row, new_container_ids):
        self.docid = row.docid
        self.namespace = row.namespace
        self.name = row.name
        self.deleted_time = row.deleted_time
        self.deleted_by = row.deleted_by
        self.new_container_ids = new_container_ids




from sqlalchemy.types import TypeDecorator
from sqlalchemy.types import UnicodeText
import simplejson as json


class JSONType(TypeDecorator):
    """Holds a JSON-serialized object.

    JSONType builds upon the UnicodeText type to apply Python's
    ``json.dumps()`` to incoming objects, and ``json.loads()`` on
    the way out, allowing any JSON compatible object to be stored as
    a Unicode text field.

    Note: This column type is immutable, meaning SQLAlchemy
    will not notice if something changes the content of values
    read from columns of this type.  Values should be replaced only.
    """

    impl = UnicodeText

    def bind_processor(self, dialect):
        impl_processor = self.impl.bind_processor(dialect)
        use_impl_processor = bool(impl_processor)

        def process(value):
            if value is not None:
                value = json.dumps(value, separators=(',', ':'))
                if not isinstance(value, unicode):
                    value = unicode(value, 'ascii')
            if use_impl_processor:
                value = impl_processor(value)
            return value

        return process

    def result_processor(self, dialect, coltype):
        impl_processor = self.impl.result_processor(dialect, coltype)
        use_impl_processor = bool(impl_processor)

        def process(value):
            if use_impl_processor:
                value = impl_processor(value)  # pragma: no cover
            if value:
                value = json.loads(value)
            else:
                value = None
            return value

        return process




from zope.interface import Attribute
from zope.interface import Interface

try:  # pragma: no cover
    from zope.dublincore.interfaces import IDCDescriptiveProperties
    from zope.dublincore.interfaces import IDCTimes
except ImportError:
    # We don't really need to depend on the bulky zope.dublincore
    # package if it's not installed.  This is a copy of the interfaces
    # we are interested in.
    from zope.schema import Datetime
    from zope.schema import Text
    from zope.schema import TextLine


    class IDCDescriptiveProperties(Interface):
        """Basic descriptive meta-data properties
        """

        title = TextLine(
            title = u'Title',
            description =
            u"The first unqualified Dublin Core 'Title' element value."
            )

        description = Text(
            title = u'Description',
            description =
            u"The first unqualified Dublin Core 'Description' element value.",
            )


    class IDCTimes(Interface):
        """Time properties
        """

        created = Datetime(
            title = u'Creation Date',
            description =
            u"The date and time that an object is created. "
            u"\nThis is normally set automatically."
            )

        modified = Datetime(
            title = u'Modification Date',
            description =
            u"The date and time that the object was last modified in a\n"
            u"meaningful way."
            )


class IArchive(Interface):
    """A utility that archives objects."""

    def archive(obj):
        """Add a version to the archive of an object.

        The obj parameter must provide the IObjectVersion interface.
        The object does not need to have been in the archive previously.

        Returns the new version number.
        """

    def history(docid, only_current=False):
        """Get the history of an object.

        Returns a list of objects that provide IObjectHistoryRecord.
        The most recent version number is listed first.

        If only_current is true, only the current history record
        is returned in the list.  (The most current history record
        might not be the most recent version number if the object
        has been reverted.)
        """

    def get_version(docid, version_num):
        """Return a specific IObjectHistoryRecord for a document.
        """

    def reverted(docid, version_num):
        """Tell the database that an object has been reverted."""

    def archive_container(container, user):
        """Update the archive of a container.

        The container parameter must provide the IContainerVersion interface.
        The container does not need to have been in the archive previously.
        The user parameter (a string) specifies who made the change.

        Note that this method does not keep a record of container versions.
        It only records the contents of the container and tracks deletions,
        to allow for undeletion of contained objects.

        Returns None.
        """

    def container_contents(container_id):
        """Returns the contents of a container as IContainerRecord.
        """


class IObjectVersion(IDCDescriptiveProperties, IDCTimes):
    """The content of an object for version control.

    Note that the following attributes are required, as specified by
    the base interfaces:

        title
        description
        created
        modified

    The title and description should be unicode.
    """

    docid = Attribute("The docid of the object as an integer.")

    path = Attribute(
        """The path of the object as a Unicode string.

        Used only as metadata.  May be left empty.
        """)

    attrs = Attribute(
        """The attributes to store as a JSON-encodeable dictionary.

        May be None.
        """)

    blobs = Attribute(
        """A map of binary large objects linked to this state.  May be None.

        Each key is a unicode string.  Each value is either a
        filename or an open file object (such as a StringIO).
        Open file objects must be seekable.
        """)

    klass = Attribute(
        """Optional: the class of the object.

        To detect the class automatically, do not provide this attribute,
        or set it to None.
        """)

    user = Attribute("The user who made the change. (A string)")

    comment = Attribute("The comment linked to the version; may be None.")


class IObjectHistoryRecord(IObjectVersion):
    """An historical record of an object version.

    The IArchive.history() method returns objects that provide this
    interface.  All blobs returned by the history() method are open
    files (not filenames).
    """

    version_num = Attribute("The version number of the object; starts with 1.")

    derived_from_version = Attribute(
        """The version number this version was based on.

        None if the object was created in this version.
        """)

    current_version = Attribute("The current version number of the object.")

    archive_time = Attribute(
        """The datetime in UTC when the version was archived.

        May be different from the modified attribute.  The modified
        attribute is set by the application, whereas the archive_time
        attribute is set by repozitory automatically.
        """)


class IContainerVersion(Interface):
    """The contents of a container for version control."""

    container_id = Attribute("The ID of the container as an integer.")

    path = Attribute(
        """The path of the container as a Unicode string.

        Used only as metadata.  May be left empty.
        """)

    map = Attribute(
        """The current items in the container, as {name: docid}.

        All names must be non-empty strings and all referenced objects
        must already exist in the archive.
        """)

    ns_map = Attribute(
        """Namespaced container items, as {ns: {name: docid}}.

        All namespaces and names must be non-empty strings and all referenced
        objects must already exist in the archive.
        """)


class IContainerRecord(IContainerVersion):
    """Provides the current and deleted contents of a container."""

    deleted = Attribute(
        """The deleted items in the container as a list of IDeletedItem.

        The most recently deleted items are listed first.

        A item name may appear more than once in the list if it has referred
        to different docids at different times.  A docid will never appear
        more than once in the list, since adding an object to a container
        causes the corresponding docid to be removed from the deleted list
        for that container.
        """)


class IDeletedItem(Interface):
    """A record of an item deleted from a container."""

    docid = Attribute("The docid of the deleted object as an integer.")

    namespace = Attribute("""The object's former namespace in the container.

    Empty if the object was not in a namespace.
    """)

    name = Attribute("The object's former name within the container.")

    deleted_time = Attribute("When the object was deleted (a UTC datetime).")

    deleted_by = Attribute("Who deleted the object (a string).")

    new_container_ids = Attribute("""Container(s) where the object now exists.

    Empty or None if the object is not currently in any container.
    """)



"""Tests of repozitory.archive"""

from StringIO import StringIO
import datetime
import unittest2 as unittest


class ArchiveTest(unittest.TestCase):

    def setUp(self):
        import transaction
        transaction.abort()
        from repozitory.archive import forget_sessions
        forget_sessions()

    def tearDown(self):
        import transaction
        transaction.abort()
        from repozitory.archive import forget_sessions
        forget_sessions()

    def _class(self):
        from repozitory.archive import Archive
        return Archive

    def _make(self, *args, **kw):
        return self._class()(*args, **kw)

    def _make_default(self):
        from repozitory.archive import EngineParams
        params = EngineParams('sqlite:///')
        return self._make(params)

    def _make_dummy_object_version(self):
        return DummyObjectVersion()

    def test_verifyImplements_IArchive(self):
        from zope.interface.verify import verifyClass
        from repozitory.interfaces import IArchive
        verifyClass(IArchive, self._class())

    def test_verifyProvides_IArchive(self):
        from zope.interface.verify import verifyObject
        from repozitory.interfaces import IArchive
        verifyObject(IArchive, self._make_default())

    def test_query_session_with_empty_database(self):
        from repozitory.schema import ArchivedObject
        archive = self._make_default()
        session = archive.session
        q = session.query(ArchivedObject).count()
        self.assertEqual(q, 0)

    def test_archive_simple_object(self):
        obj = self._make_dummy_object_version()
        archive = self._make_default()
        ver = archive.archive(obj)
        self.assertEqual(ver, 1)

        from repozitory.schema import ArchivedObject
        rows = archive.session.query(ArchivedObject).all()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0].docid, 4)
        self.assertEqual(rows[0].created, datetime.datetime(2011, 4, 6))

        from repozitory.schema import ArchivedClass
        rows = archive.session.query(ArchivedClass).all()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0].module, u'repozitory.tests.test_archive')
        self.assertEqual(rows[0].name, u'DummyObjectVersion')

        from repozitory.schema import ArchivedState
        rows = archive.session.query(ArchivedState).all()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0].docid, 4)
        self.assertEqual(rows[0].version_num, 1)
        self.assertEqual(rows[0].path, u'/my/object')
        self.assertEqual(rows[0].modified, datetime.datetime(2011, 4, 7))
        self.assertEqual(rows[0].title, u'Cool Object')
        self.assertEqual(rows[0].description, None)
        self.assertEqual(rows[0].attrs, {'a': 1, 'b': [2]})
        self.assertEqual(rows[0].comment, u'I like version control.')

        from repozitory.schema import ArchivedCurrent
        rows = archive.session.query(ArchivedCurrent).all()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0].docid, 4)
        self.assertEqual(rows[0].version_num, 1)

        from repozitory.schema import ArchivedBlobInfo
        rows = archive.session.query(ArchivedBlobInfo).all()
        self.assertEqual(len(rows), 0)

        from repozitory.schema import ArchivedChunk
        rows = archive.session.query(ArchivedChunk).all()
        self.assertEqual(len(rows), 0)

        from repozitory.schema import ArchivedBlobLink
        rows = archive.session.query(ArchivedBlobLink).all()
        self.assertEqual(len(rows), 0)

        from repozitory.schema import ArchivedContainer
        rows = archive.session.query(ArchivedContainer).all()
        self.assertEqual(len(rows), 0)

        from repozitory.schema import ArchivedItem
        rows = archive.session.query(ArchivedItem).all()
        self.assertEqual(len(rows), 0)

        from repozitory.schema import ArchivedItemDeleted
        rows = archive.session.query(ArchivedItemDeleted).all()
        self.assertEqual(len(rows), 0)

    def test_archive_object_with_no_attrs(self):
        obj = self._make_dummy_object_version()
        obj.attrs = None
        archive = self._make_default()
        ver = archive.archive(obj)
        self.assertEqual(ver, 1)

        from repozitory.schema import ArchivedState
        rows = archive.session.query(ArchivedState).all()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0].docid, 4)
        self.assertEqual(rows[0].version_num, 1)
        self.assertEqual(rows[0].path, u'/my/object')
        self.assertEqual(rows[0].modified, datetime.datetime(2011, 4, 7))
        self.assertEqual(rows[0].title, u'Cool Object')
        self.assertEqual(rows[0].description, None)
        self.assertEqual(rows[0].attrs, None)
        self.assertEqual(rows[0].comment, u'I like version control.')

    def test_archive_2_revisions_of_simple_object(self):
        obj = self._make_dummy_object_version()
        archive = self._make_default()
        v1 = archive.archive(obj)
        self.assertEqual(v1, 1)

        obj.title = 'New Title!'
        obj.comment = 'I still like version control.'
        v2 = archive.archive(obj)
        self.assertEqual(v2, 2)

        from repozitory.schema import ArchivedObject
        rows = archive.session.query(ArchivedObject).all()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0].docid, 4)
        self.assertEqual(rows[0].created, datetime.datetime(2011, 4, 6))

        from repozitory.schema import ArchivedClass
        rows = archive.session.query(ArchivedClass).all()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0].module, u'repozitory.tests.test_archive')
        self.assertEqual(rows[0].name, u'DummyObjectVersion')

        from repozitory.schema import ArchivedState
        rows = (archive.session.query(ArchivedState)
            .order_by(ArchivedState.version_num)
            .all())
        self.assertEqual(len(rows), 2)
        self.assertEqual(rows[0].docid, 4)
        self.assertEqual(rows[0].version_num, 1)
        self.assertEqual(rows[0].path, u'/my/object')
        self.assertEqual(rows[0].modified, datetime.datetime(2011, 4, 7))
        self.assertEqual(rows[0].title, u'Cool Object')
        self.assertEqual(rows[0].description, None)
        self.assertEqual(rows[0].attrs, {'a': 1, 'b': [2]})
        self.assertEqual(rows[0].comment, u'I like version control.')
        self.assertEqual(rows[1].docid, 4)
        self.assertEqual(rows[1].version_num, 2)
        self.assertEqual(rows[1].path, u'/my/object')
        self.assertEqual(rows[1].modified, datetime.datetime(2011, 4, 7))
        self.assertEqual(rows[1].title, u'New Title!')
        self.assertEqual(rows[1].description, None)
        self.assertEqual(rows[1].attrs, {'a': 1, 'b': [2]})
        self.assertEqual(rows[1].comment, u'I still like version control.')

        from repozitory.schema import ArchivedCurrent
        rows = archive.session.query(ArchivedCurrent).all()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0].docid, 4)
        self.assertEqual(rows[0].version_num, 2)

    def test_archive_with_blob(self):
        obj = self._make_dummy_object_version()
        obj.blobs = {'readme.txt': StringIO('42')}
        archive = self._make_default()
        archive.archive(obj)

        from repozitory.schema import ArchivedBlobInfo
        rows = archive.session.query(ArchivedBlobInfo).all()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0].chunk_count, 1)
        self.assertEqual(rows[0].length, 2)
        self.assertEqual(rows[0].md5,
            'a1d0c6e83f027327d8461063f4ac58a6')
        self.assertEqual(rows[0].sha256,
            '73475cb40a568e8da8a045ced110137e159f890ac4da883b6b17dc651b3a8049')

        from repozitory.schema import ArchivedChunk
        rows = archive.session.query(ArchivedChunk).all()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0].chunk_index, 0)
        self.assertEqual(rows[0].data, '42')

        from repozitory.schema import ArchivedBlobLink
        rows = archive.session.query(ArchivedBlobLink).all()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0].docid, 4)
        self.assertEqual(rows[0].version_num, 1)
        self.assertEqual(rows[0].name, 'readme.txt')

    def test_archive_with_filename_blob(self):
        import tempfile
        f = tempfile.NamedTemporaryFile()
        f.write('42')
        f.flush()

        obj = self._make_dummy_object_version()
        obj.blobs = {'readme.txt': f.name}
        archive = self._make_default()
        archive.archive(obj)

        from repozitory.schema import ArchivedChunk
        rows = archive.session.query(ArchivedChunk).all()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0].chunk_index, 0)
        self.assertEqual(rows[0].data, '42')

        from repozitory.schema import ArchivedBlobLink
        rows = archive.session.query(ArchivedBlobLink).all()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0].docid, 4)
        self.assertEqual(rows[0].version_num, 1)
        self.assertEqual(rows[0].name, 'readme.txt')

    def test_archive_deduplicates_blobs(self):
        obj = self._make_dummy_object_version()
        obj.blobs = {'readme.txt': StringIO('42')}
        archive = self._make_default()
        archive.archive(obj)
        obj.blobs['readme2.txt'] = StringIO('24.')
        archive.archive(obj)

        from repozitory.schema import ArchivedBlobInfo
        rows = (archive.session.query(ArchivedBlobInfo)
            .order_by(ArchivedBlobInfo.length)
            .all())
        self.assertEqual(len(rows), 2)
        self.assertEqual(rows[0].length, 2)
        self.assertEqual(rows[1].length, 3)

        from repozitory.schema import ArchivedChunk
        from sqlalchemy import func
        rows = (archive.session.query(ArchivedChunk)
            .order_by(func.length(ArchivedChunk.data))
            .all())
        self.assertEqual(len(rows), 2)
        self.assertEqual(rows[0].chunk_index, 0)
        self.assertEqual(rows[0].data, '42')
        self.assertEqual(rows[1].chunk_index, 0)
        self.assertEqual(rows[1].data, '24.')

        from repozitory.schema import ArchivedBlobLink
        rows = (archive.session.query(ArchivedBlobLink)
            .order_by(ArchivedBlobLink.version_num, ArchivedBlobLink.name)
            .all())
        self.assertEqual(len(rows), 3)
        self.assertEqual(rows[0].docid, 4)
        self.assertEqual(rows[0].version_num, 1)
        self.assertEqual(rows[0].name, 'readme.txt')
        self.assertEqual(rows[1].docid, 4)
        self.assertEqual(rows[1].version_num, 2)
        self.assertEqual(rows[1].name, 'readme.txt')
        self.assertEqual(rows[2].docid, 4)
        self.assertEqual(rows[2].version_num, 2)
        self.assertEqual(rows[2].name, 'readme2.txt')

        # Confirm that rows 0 and 1, which are different versions,
        # refer to the same blob_id.
        self.assertEqual(rows[0].blob_id, rows[1].blob_id)
        # Row 2 refers to a different blob.
        self.assertNotEqual(rows[0].blob_id, rows[2].blob_id)

    def test_archive_broken_class(self):
        class Unpickleable:
            # Instances of this class should not be stored because the
            # class can not be resolved through a module and name reference.
            pass

        obj = self._make_dummy_object_version()
        obj.klass = Unpickleable
        archive = self._make_default()
        with self.assertRaises(TypeError):
            archive.archive(obj)

    def test_history_item_implements_IObjectHistoryRecord(self):
        obj = self._make_dummy_object_version()
        obj.comment = 'change 1'
        archive = self._make_default()
        archive.archive(obj)

        records = archive.history(obj.docid)
        self.assertEqual(len(records), 1)

        from zope.interface.verify import verifyObject
        from repozitory.interfaces import IObjectHistoryRecord
        verifyObject(IObjectHistoryRecord, records[0])

    def test_history_without_blobs(self):
        obj = self._make_dummy_object_version()
        obj.comment = 'change 1'
        archive = self._make_default()
        archive.archive(obj)
        obj.title = 'Changed Title'
        obj.description = 'New Description'
        obj.modified = datetime.datetime(2011, 4, 11)
        obj.user = 'mixer upper'
        obj.comment = None
        archive.archive(obj)

        records = archive.history(obj.docid)
        self.assertEqual(len(records), 2)

        self.assertEqual(records[0].docid, 4)
        self.assertEqual(records[0].path, u'/my/object')
        self.assertEqual(records[0].created, datetime.datetime(2011, 4, 6))
        self.assertEqual(records[0].modified, datetime.datetime(2011, 4, 11))
        self.assertEqual(records[0].version_num, 2)
        self.assertEqual(records[0].current_version, 2)
        self.assertEqual(records[0].title, u'Changed Title')
        self.assertEqual(records[0].description, u'New Description')
        self.assertEqual(records[0].attrs, {'a': 1, 'b': [2]})
        self.assertEqual(records[0].user, 'mixer upper')
        self.assertEqual(records[0].comment, None)
        self.assertFalse(records[0].blobs)
        self.assertEqual(records[0].klass, DummyObjectVersion)

        self.assertEqual(records[1].docid, 4)
        self.assertEqual(records[1].path, u'/my/object')
        self.assertEqual(records[1].created, datetime.datetime(2011, 4, 6))
        self.assertEqual(records[1].modified, datetime.datetime(2011, 4, 7))
        self.assertEqual(records[1].version_num, 1)
        self.assertEqual(records[1].current_version, 2)
        self.assertEqual(records[1].title, u'Cool Object')
        self.assertEqual(records[1].description, None)
        self.assertEqual(records[1].attrs, {'a': 1, 'b': [2]})
        self.assertEqual(records[1].user, 'tester')
        self.assertEqual(records[1].comment, 'change 1')
        self.assertFalse(records[1].blobs)
        self.assertEqual(records[1].klass, DummyObjectVersion)

        self.assertGreater(records[0].archive_time, records[0].created)

    def test_history_with_small_blob(self):
        obj = self._make_dummy_object_version()
        archive = self._make_default()
        archive.archive(obj)
        obj.blobs = {'x': StringIO('42')}
        archive.archive(obj)

        from repozitory.schema import ArchivedChunk
        rows = archive.session.query(ArchivedChunk).all()
        self.assertEqual(len(rows), 1)

        records = archive.history(obj.docid)
        self.assertEqual(len(records), 2)

        self.assertTrue(records[0].blobs)
        self.assertEqual(records[0].blobs.keys(), ['x'])
        blob = records[0].blobs['x']
        self.assertEqual(blob.read(), '42')
        with self.assertRaises(IOError):
            blob.write('x')
        with self.assertRaises(IOError):
            blob.writelines(['x'])

        self.assertFalse(records[1].blobs)

    def test_history_with_large_blob(self):
        archive = self._make_default()
        obj = self._make_dummy_object_version()
        obj.blobs = {'x': StringIO('*' * 10485760)}  # 10 MiB
        archive.archive(obj)

        from repozitory.schema import ArchivedChunk
        rows = archive.session.query(ArchivedChunk).all()
        self.assertEqual(len(rows), 10)

        records = archive.history(obj.docid)
        self.assertEqual(len(records), 1)
        self.assertEqual(records[0].blobs.keys(), ['x'])
        blob = records[0].blobs['x']
        self.assertEqual(len(blob.read()), 10485760)

    def test_history_only_current(self):
        archive = self._make_default()
        obj = self._make_dummy_object_version()
        archive.archive(obj)
        obj.attrs = None
        archive.archive(obj)

        records = archive.history(obj.docid, only_current=True)
        self.assertEqual(len(records), 1)
        self.assertEqual(records[0].attrs, {})


    def test_get_version(self):
        archive = self._make_default()
        obj = self._make_dummy_object_version()
        archive.archive(obj)
        obj.attrs = None
        archive.archive(obj)

        v1 = archive.get_version(obj.docid, 1)
        v2 = archive.get_version(obj.docid, 2)
        self.assertEqual(v1.attrs, {u'a': 1, u'b': [2]})
        self.assertEqual(v2.attrs, {})
        self.assertEqual(v1.version_num, 1)
        self.assertEqual(v2.version_num, 2)
        self.assertEqual(v1.current_version, 2)
        self.assertEqual(v2.current_version, 2)

    def test_reverted(self):
        obj = self._make_dummy_object_version()
        archive = self._make_default()
        v1 = archive.archive(obj)
        self.assertEqual(v1, 1)

        obj.title = 'New Title!'
        v2 = archive.archive(obj)
        self.assertEqual(v2, 2)

        records = archive.history(obj.docid)
        self.assertEqual(len(records), 2)
        self.assertEqual(records[0].version_num, 2)
        self.assertEqual(records[1].version_num, 1)
        self.assertEqual(records[0].current_version, 2)
        self.assertEqual(records[1].current_version, 2)

        archive.reverted(obj.docid, 1)

        records = archive.history(obj.docid)
        self.assertEqual(len(records), 2)
        self.assertEqual(records[0].version_num, 2)
        self.assertEqual(records[1].version_num, 1)
        self.assertEqual(records[0].current_version, 1)
        self.assertEqual(records[1].current_version, 1)

    def test_archive_container_empty(self):
        archive = self._make_default()

        class DummyContainerVersion:
            container_id = 5
            path = '/my/container'
            map = {}
            ns_map = {}

        archive.archive_container(DummyContainerVersion(), 'testuser')

        from repozitory.schema import ArchivedContainer
        rows = archive.session.query(ArchivedContainer).all()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0].container_id, 5)
        self.assertEqual(rows[0].path, u'/my/container')

        from repozitory.schema import ArchivedItem
        rows = (archive.session.query(ArchivedItem)
            .order_by(ArchivedItem.namespace).all())
        self.assertEqual(len(rows), 0)

        from repozitory.schema import ArchivedItemDeleted
        rows = archive.session.query(ArchivedItemDeleted).all()
        self.assertEqual(len(rows), 0)

    def test_archive_container_non_empty(self):
        archive = self._make_default()
        obj4 = self._make_dummy_object_version()
        obj6 = self._make_dummy_object_version()
        obj6.docid = 6
        archive.archive(obj4)
        archive.archive(obj6)

        class DummyContainerVersion:
            container_id = 5
            path = '/my/container'
            map = {'a': 4}
            ns_map = {'headers': {'b': 6}}

        archive.archive_container(DummyContainerVersion(), 'testuser')

        from repozitory.schema import ArchivedContainer
        rows = archive.session.query(ArchivedContainer).all()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0].container_id, 5)
        self.assertEqual(rows[0].path, u'/my/container')

        from repozitory.schema import ArchivedItem
        rows = (archive.session.query(ArchivedItem)
            .order_by(ArchivedItem.namespace).all())
        self.assertEqual(len(rows), 2)
        self.assertEqual(rows[0].container_id, 5)
        self.assertEqual(rows[0].namespace, u'')
        self.assertEqual(rows[0].name, u'a')
        self.assertEqual(rows[0].docid, 4)
        self.assertEqual(rows[1].container_id, 5)
        self.assertEqual(rows[1].namespace, u'headers')
        self.assertEqual(rows[1].name, u'b')
        self.assertEqual(rows[1].docid, 6)

        from repozitory.schema import ArchivedItemDeleted
        rows = archive.session.query(ArchivedItemDeleted).all()
        self.assertEqual(len(rows), 0)

    def test_archive_container_with_deletion(self):
        archive = self._make_default()
        obj4 = self._make_dummy_object_version()
        obj6 = self._make_dummy_object_version()
        obj6.docid = 6
        archive.archive(obj4)
        archive.archive(obj6)

        class DummyContainerVersion:
            container_id = 5
            path = '/my/container'
            map = {'a': 4}
            ns_map = {'headers': {'b': 6}}

        c = DummyContainerVersion()
        archive.archive_container(c, 'user1')
        c.ns_map = {}
        archive.archive_container(c, 'user2')

        from repozitory.schema import ArchivedContainer
        rows = archive.session.query(ArchivedContainer).all()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0].container_id, 5)
        self.assertEqual(rows[0].path, u'/my/container')

        from repozitory.schema import ArchivedItem
        rows = archive.session.query(ArchivedItem).all()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0].container_id, 5)
        self.assertEqual(rows[0].namespace, u'')
        self.assertEqual(rows[0].name, u'a')
        self.assertEqual(rows[0].docid, 4)

        from repozitory.schema import ArchivedItemDeleted
        rows = archive.session.query(ArchivedItemDeleted).all()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0].container_id, 5)
        self.assertEqual(rows[0].namespace, u'headers')
        self.assertEqual(rows[0].name, u'b')
        self.assertEqual(rows[0].docid, 6)
        self.assertEqual(rows[0].deleted_by, 'user2')
        self.assertTrue(rows[0].deleted_time)

    def test_archive_container_with_undeletion(self):
        archive = self._make_default()
        obj4 = self._make_dummy_object_version()
        obj6 = self._make_dummy_object_version()
        obj6.docid = 6
        archive.archive(obj4)
        archive.archive(obj6)

        class DummyContainerVersion:
            container_id = 5
            path = '/my/container'
            map = {'a': 4}
            ns_map = {'headers': {'b': 6}}

        c = DummyContainerVersion()
        archive.archive_container(c, 'user1')
        c.ns_map = {}
        archive.archive_container(c, 'user2')
        c.ns_map = {'headers': {'z': 6}}
        archive.archive_container(c, 'user3')

        from repozitory.schema import ArchivedContainer
        rows = archive.session.query(ArchivedContainer).all()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0].container_id, 5)
        self.assertEqual(rows[0].path, u'/my/container')

        from repozitory.schema import ArchivedItem
        rows = (archive.session.query(ArchivedItem)
            .order_by(ArchivedItem.namespace).all())
        self.assertEqual(len(rows), 2)
        self.assertEqual(rows[0].container_id, 5)
        self.assertEqual(rows[0].namespace, u'')
        self.assertEqual(rows[0].name, u'a')
        self.assertEqual(rows[0].docid, 4)
        self.assertEqual(rows[1].container_id, 5)
        self.assertEqual(rows[1].namespace, u'headers')
        self.assertEqual(rows[1].name, u'z')
        self.assertEqual(rows[1].docid, 6)

        from repozitory.schema import ArchivedItemDeleted
        rows = archive.session.query(ArchivedItemDeleted).all()
        self.assertEqual(len(rows), 0)

    def test_archive_container_with_no_change(self):
        archive = self._make_default()
        obj4 = self._make_dummy_object_version()
        obj6 = self._make_dummy_object_version()
        obj6.docid = 6
        archive.archive(obj4)
        archive.archive(obj6)

        class DummyContainerVersion:
            container_id = 5
            path = '/my/container'
            map = {'a': 4}
            ns_map = {'headers': {'b': 6}}

        c = DummyContainerVersion()
        archive.archive_container(c, 'user1')
        archive.archive_container(c, 'user2')

        from repozitory.schema import ArchivedContainer
        rows = archive.session.query(ArchivedContainer).all()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0].container_id, 5)
        self.assertEqual(rows[0].path, u'/my/container')

        from repozitory.schema import ArchivedItem
        rows = (archive.session.query(ArchivedItem)
            .order_by(ArchivedItem.namespace).all())
        self.assertEqual(len(rows), 2)
        self.assertEqual(rows[0].container_id, 5)
        self.assertEqual(rows[0].namespace, u'')
        self.assertEqual(rows[0].name, u'a')
        self.assertEqual(rows[0].docid, 4)
        self.assertEqual(rows[1].container_id, 5)
        self.assertEqual(rows[1].namespace, u'headers')
        self.assertEqual(rows[1].name, u'b')
        self.assertEqual(rows[1].docid, 6)

        from repozitory.schema import ArchivedItemDeleted
        rows = archive.session.query(ArchivedItemDeleted).all()
        self.assertEqual(len(rows), 0)

    def test_archive_container_with_item_rename(self):
        archive = self._make_default()
        obj4 = self._make_dummy_object_version()
        obj6 = self._make_dummy_object_version()
        obj6.docid = 6
        archive.archive(obj4)
        archive.archive(obj6)

        class DummyContainerVersion:
            container_id = 5
            path = '/my/container'
            map = {'a': 4}
            ns_map = {'headers': {'b': 6}}

        c = DummyContainerVersion()
        archive.archive_container(c, 'user1')
        c.map = {'z': 4}
        archive.archive_container(c, 'user2')

        from repozitory.schema import ArchivedContainer
        rows = archive.session.query(ArchivedContainer).all()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0].container_id, 5)
        self.assertEqual(rows[0].path, u'/my/container')

        from repozitory.schema import ArchivedItem
        rows = (archive.session.query(ArchivedItem)
            .order_by(ArchivedItem.namespace).all())
        self.assertEqual(len(rows), 2)
        self.assertEqual(rows[0].container_id, 5)
        self.assertEqual(rows[0].namespace, u'')
        self.assertEqual(rows[0].name, u'z')
        self.assertEqual(rows[0].docid, 4)
        self.assertEqual(rows[1].container_id, 5)
        self.assertEqual(rows[1].namespace, u'headers')
        self.assertEqual(rows[1].name, u'b')
        self.assertEqual(rows[1].docid, 6)

        from repozitory.schema import ArchivedItemDeleted
        rows = archive.session.query(ArchivedItemDeleted).all()
        self.assertEqual(len(rows), 0)

    def test_archive_container_with_path_change(self):
        archive = self._make_default()
        obj4 = self._make_dummy_object_version()
        archive.archive(obj4)

        class DummyContainerVersion:
            container_id = 5
            path = '/my/container'
            map = {'a': 4}
            ns_map = None

        c = DummyContainerVersion()
        archive.archive_container(c, 'user1')
        c.path = '/your/container'
        archive.archive_container(c, 'user2')

        from repozitory.schema import ArchivedContainer
        rows = archive.session.query(ArchivedContainer).all()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0].container_id, 5)
        self.assertEqual(rows[0].path, u'/your/container')

        from repozitory.schema import ArchivedItem
        rows = (archive.session.query(ArchivedItem)
            .order_by(ArchivedItem.namespace).all())
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0].container_id, 5)
        self.assertEqual(rows[0].namespace, u'')
        self.assertEqual(rows[0].name, u'a')
        self.assertEqual(rows[0].docid, 4)

        from repozitory.schema import ArchivedItemDeleted
        rows = archive.session.query(ArchivedItemDeleted).all()
        self.assertEqual(len(rows), 0)

    def test_archive_container_with_changing_docid(self):
        archive = self._make_default()
        obj4 = self._make_dummy_object_version()
        obj6 = self._make_dummy_object_version()
        obj6.docid = 6
        archive.archive(obj4)
        archive.archive(obj6)

        class DummyContainerVersion:
            container_id = 5
            path = '/my/container'
            map = {'a': 4}
            ns_map = {'headers': {'b': 6}}

        c = DummyContainerVersion()
        archive.archive_container(c, 'user1')
        c.map = {'a': 6}
        archive.archive_container(c, 'user2')

        from repozitory.schema import ArchivedContainer
        rows = archive.session.query(ArchivedContainer).all()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0].container_id, 5)
        self.assertEqual(rows[0].path, u'/my/container')

        from repozitory.schema import ArchivedItem
        rows = (archive.session.query(ArchivedItem)
            .order_by(ArchivedItem.namespace).all())
        self.assertEqual(len(rows), 2)
        self.assertEqual(rows[0].container_id, 5)
        self.assertEqual(rows[0].namespace, u'')
        self.assertEqual(rows[0].name, u'a')
        self.assertEqual(rows[0].docid, 6)
        self.assertEqual(rows[1].container_id, 5)
        self.assertEqual(rows[1].namespace, u'headers')
        self.assertEqual(rows[1].name, u'b')
        self.assertEqual(rows[1].docid, 6)

        from repozitory.schema import ArchivedItemDeleted
        rows = archive.session.query(ArchivedItemDeleted).all()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0].container_id, 5)
        self.assertEqual(rows[0].namespace, u'')
        self.assertEqual(rows[0].name, u'a')
        self.assertEqual(rows[0].docid, 4)
        self.assertEqual(rows[0].deleted_by, 'user2')
        self.assertTrue(rows[0].deleted_time)

    def test_container_contents_empty(self):
        archive = self._make_default()

        class DummyContainerVersion:
            container_id = 5
            path = '/my/container'
            map = {}
            ns_map = {}

        archive.archive_container(DummyContainerVersion(), 'testuser')

        r = archive.container_contents(5)
        self.assertEqual(r.container_id, 5)
        self.assertEqual(r.path, u'/my/container')
        self.assertEqual(r.map, {})
        self.assertEqual(r.ns_map, {})
        self.assertEqual(r.deleted, [])

    def test_container_contents_non_empty(self):
        archive = self._make_default()
        obj4 = self._make_dummy_object_version()
        obj6 = self._make_dummy_object_version()
        obj6.docid = 6
        archive.archive(obj4)
        archive.archive(obj6)

        class DummyContainerVersion:
            container_id = 5
            path = '/my/container'
            map = {'a': 4}
            ns_map = {'headers': {'b': 6}}

        archive.archive_container(DummyContainerVersion(), 'testuser')

        r = archive.container_contents(5)

        from zope.interface.verify import verifyObject
        from repozitory.interfaces import IContainerRecord
        verifyObject(IContainerRecord, r)

        self.assertEqual(r.container_id, 5)
        self.assertEqual(r.path, u'/my/container')
        self.assertEqual(r.map, {'a': 4})
        self.assertEqual(r.ns_map, {'headers': {'b': 6}})
        self.assertEqual(r.deleted, [])

    def test_container_contents_with_deletion(self):
        archive = self._make_default()
        obj4 = self._make_dummy_object_version()
        obj6 = self._make_dummy_object_version()
        obj6.docid = 6
        archive.archive(obj4)
        archive.archive(obj6)

        class DummyContainerVersion:
            container_id = 5
            path = '/my/container'
            map = {'a': 4}
            ns_map = {'headers': {'b': 6}}

        c = DummyContainerVersion()
        archive.archive_container(c, 'user1')
        c.ns_map = {}
        archive.archive_container(c, 'user2')

        r = archive.container_contents(5)
        self.assertEqual(r.container_id, 5)
        self.assertEqual(r.path, u'/my/container')
        self.assertEqual(r.map, {'a': 4})
        self.assertEqual(r.ns_map, {})
        self.assertEqual(len(r.deleted), 1)

        row = r.deleted[0]
        from zope.interface.verify import verifyObject
        from repozitory.interfaces import IDeletedItem
        verifyObject(IDeletedItem, row)

        self.assertEqual(r.deleted[0].docid, 6)
        self.assertEqual(r.deleted[0].namespace, 'headers')
        self.assertEqual(r.deleted[0].name, 'b')
        self.assertEqual(r.deleted[0].deleted_by, 'user2')
        self.assertTrue(r.deleted[0].deleted_time)
        self.assertFalse(r.deleted[0].new_container_ids)

    def test_container_contents_after_move(self):
        archive = self._make_default()
        obj4 = self._make_dummy_object_version()
        archive.archive(obj4)

        class DummyContainerVersion:
            def __init__(self, container_id, path):
                self.container_id = container_id
                self.path = path
                self.map = {}
                self.ns_map = {}

        # user1 creates two containers, c5 and c6, putting obj4 in c5.
        c5 = DummyContainerVersion(5, '/c5')
        c5.map = {'a': 4}
        archive.archive_container(c5, 'user1')
        c6 = DummyContainerVersion(6, '/c6')
        archive.archive_container(c6, 'user1')

        # user2 moves obj4 from c5 to c6.
        c5.map = {}
        archive.archive_container(c5, 'user2')
        c6.map = {'a': 4}
        archive.archive_container(c6, 'user2')

        # List the contents of c5.  It should have a deletion record
        # with new_container_ids providing the new location of the object.
        r = archive.container_contents(5)
        self.assertEqual(r.container_id, 5)
        self.assertEqual(r.map, {})
        self.assertEqual(r.ns_map, {})
        self.assertEqual(len(r.deleted), 1)

        row = r.deleted[0]
        from zope.interface.verify import verifyObject
        from repozitory.interfaces import IDeletedItem
        verifyObject(IDeletedItem, row)

        self.assertEqual(r.deleted[0].docid, 4)
        self.assertEqual(r.deleted[0].namespace, '')
        self.assertEqual(r.deleted[0].name, 'a')
        self.assertEqual(r.deleted[0].deleted_by, 'user2')
        self.assertTrue(r.deleted[0].deleted_time)
        self.assertEqual(r.deleted[0].new_container_ids, [6])


class DummyObjectVersion:
    docid = 4
    path = '/my/object'
    created = datetime.datetime(2011, 4, 6)
    modified = datetime.datetime(2011, 4, 7)
    title = 'Cool Object'
    description = None
    attrs = {'a': 1, 'b': [2]}
    user = 'tester'
    comment = 'I like version control.'




from repozitory.jsontype import JSONType
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import deferred
from sqlalchemy.orm import relationship
from sqlalchemy.schema import Column
from sqlalchemy.schema import ForeignKey
from sqlalchemy.schema import ForeignKeyConstraint
from sqlalchemy.types import BigInteger
from sqlalchemy.types import DateTime
from sqlalchemy.types import Integer
from sqlalchemy.types import LargeBinary
from sqlalchemy.types import String
from sqlalchemy.types import Unicode

Base = declarative_base()


class ArchivedObject(Base):
    """An object in the archive."""
    __tablename__ = 'archived_object'
    docid = Column(BigInteger, primary_key=True, nullable=False,
        autoincrement=False)
    created = Column(DateTime, nullable=False, index=True)


class ArchivedClass(Base):
    """The class of some objects in the archive."""
    __tablename__ = 'archived_class'
    class_id = Column(Integer, nullable=False, primary_key=True)
    module = Column(Unicode, nullable=False, index=True)
    name = Column(Unicode, nullable=False)


class ArchivedState(Base):
    """The state of an object in a particular version.

    Also contains info about the version commit: archive_time, user,
    and comment.
    """
    __tablename__ = 'archived_state'
    docid = Column(BigInteger, ForeignKey('archived_object.docid'),
        primary_key=True, nullable=False)
    version_num = Column(Integer, primary_key=True, nullable=False,
        autoincrement=False)
    derived_from_version = Column(Integer, nullable=True)
    class_id = Column(Integer, ForeignKey('archived_class.class_id'),
        nullable=False, index=True)
    path = Column(Unicode, nullable=False, index=True)
    modified = Column(DateTime, nullable=False, index=True)
    title = Column(Unicode, nullable=True)

    # archive_time is the time in UTC when the version was archived.
    archive_time = Column(DateTime, nullable=False)
    user = Column(Unicode, nullable=False)
    comment = Column(Unicode, nullable=True)

    description = Column(Unicode, nullable=True)
    attrs = Column(JSONType, nullable=True)

    obj = relationship(ArchivedObject)
    class_ = relationship(ArchivedClass, lazy='joined')


class ArchivedCurrent(Base):
    """Reference to the current version of an object."""
    __tablename__ = 'archived_current'
    docid = Column(BigInteger, primary_key=True, nullable=False)
    version_num = Column(Integer, nullable=False)

    __table_args__ = (
        ForeignKeyConstraint(
            ['docid', 'version_num'],
            ['archived_state.docid', 'archived_state.version_num'],
        ),
        {},
    )

    state = relationship(ArchivedState)


class ArchivedBlobInfo(Base):
    """Info about a chunked blob.

    This table provides a simple Content Addressable Storage (CAS)
    to save space in the database.
    """
    __tablename__ = 'archived_blob_info'
    blob_id = Column(Integer, primary_key=True, nullable=False)
    chunk_count = Column(Integer, nullable=False)
    length = Column(BigInteger, nullable=False)
    # Blobs are matched by both MD5 and SHA-256.
    md5 = Column(String, nullable=False, index=True)
    sha256 = Column(String, nullable=False)


class ArchivedChunk(Base):
    """A chunk of some blob data."""
    __tablename__ = 'archived_chunk'
    blob_id = Column(Integer, ForeignKey('archived_blob_info.blob_id'),
        primary_key=True, nullable=False, index=True)
    chunk_index = Column(Integer, primary_key=True, nullable=False,
        autoincrement=False)
    chunk_length = Column(Integer, nullable=False)
    data = deferred(Column(LargeBinary, nullable=False))

    blob = relationship(ArchivedBlobInfo, backref='chunks',
        order_by=chunk_index)


class ArchivedBlobLink(Base):
    """A binary file linked to a version of an object."""
    __tablename__ = 'archived_blob_link'
    docid = Column(BigInteger, primary_key=True, nullable=False)
    version_num = Column(Integer, primary_key=True, nullable=False)
    name = Column(Unicode, primary_key=True, nullable=False)
    blob_id = Column(Integer, ForeignKey('archived_blob_info.blob_id'),
        nullable=True)

    __table_args__ = (
        ForeignKeyConstraint(
            ['docid', 'version_num'],
            ['archived_state.docid', 'archived_state.version_num'],
        ),
        {},
    )

    state = relationship(ArchivedState, backref='blob_links')
    blob = relationship(ArchivedBlobInfo, lazy='joined')


class ArchivedContainer(Base):
    """A container that has version controlled objects.

    This is not the place to store data about the container itself.
    """
    __tablename__ = 'archived_container'
    container_id = Column(BigInteger, primary_key=True, nullable=False,
        autoincrement=False)
    path = Column(Unicode, nullable=False, index=True)


class ArchivedItem(Base):
    """A version controlled object in a container."""
    __tablename__ = 'archived_item'
    container_id = Column(BigInteger,
        ForeignKey('archived_container.container_id'),
        primary_key=True, index=True, nullable=False)
    namespace = Column(Unicode, primary_key=True, nullable=False,
        default=u'')
    name = Column(Unicode, primary_key=True, nullable=False)
    docid = Column(BigInteger, ForeignKey('archived_object.docid'),
        nullable=False, index=True)

    container = relationship(ArchivedContainer)
    obj = relationship(ArchivedObject)


class ArchivedItemDeleted(Base):
    """An object that has been deleted from a container.

    Does not include items that have been renamed within the container.
    """
    __tablename__ = 'archived_item_deleted'
    container_id = Column(BigInteger,
        ForeignKey('archived_container.container_id'),
        primary_key=True, index=True, nullable=False)
    docid = Column(BigInteger, ForeignKey('archived_object.docid'),
        primary_key=True, nullable=False, index=True)
    namespace = Column(Unicode, nullable=False, default=u'')
    name = Column(Unicode, nullable=False)
    deleted_time = Column(DateTime, nullable=False)
    deleted_by = Column(Unicode, nullable=False)

    container = relationship(ArchivedContainer)
    obj = relationship(ArchivedObject)




from setuptools import setup, find_packages
import os

version = '0.2'

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.txt')).read()
CHANGES = open(os.path.join(here, 'CHANGES.txt')).read()

setup(
    name='repozitory',
    version=version,
    description="Simple document versioning for web apps, "
        "especially Pyramid apps.",
    long_description=README + '\n\nChanges\n=======\n\n' + CHANGES,
    # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[],
    keywords='pyramid pylons document version versioning sql sqlalchemy',
    author='Shane Hathaway',
    author_email='shane@hathawaymix.org',
    url='https://github.com/Pylons/repozitory',
    license="BSD-derived (http://www.repoze.org/LICENSE.txt)",
    packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'psycopg2',
        'simplejson',
        'SQLAlchemy>=0.7.1',
        'zope.interface',
        'zope.schema',
        'zope.sqlalchemy',
    ],
    tests_require=['unittest2'],
    extras_require={'test': ['unittest2']},
    entry_points="""
    # -*- Entry points: -*-
    """,
)
